from __future__ import annotations
import warnings
import pickle
import torch
import numpy as np
from numpy.typing import ArrayLike
from torch._prims_common import DeviceLikeType
from torch.types import _dtype
from typing import Optional, Callable

class Dataset(torch.utils.data.Dataset):
    r"""
    Implements a torch dataset for survival data.

    Attributes
    ----------
    X (Tensor):
        Tensor with features/covariates.
    T (Tensor):
        Tensor with failure times.
    E (Tensor):
        Bool Vector with event types, 1 for failure and 0 for censoring

    Methods
    -------
    right_censor(self, censoring_time: float):
        Right censors dataset.
    save_to_file
        Saves dataset to file.
    load_from_file(cls, filename: str):
        Loads dataset from file.
    split(proportions:Arraylike, sizes: ArrayLike):
        Splits dataset into subsets according to x.
    to(...):
        Performs Tensor dtype and/or device conversion. 
    """

    def __init__(self, X: ArrayLike, T: ArrayLike, E: Optional[ArrayLike] = None, copy: bool = True) -> None:
        r"""
        Constructs a Dataset.

        Parameters
        ----------
            X : (n,...) ArrayLike
                Features/covariates with first dimension as batch dimension.
            T : (n,) ArrayLike
                Recorded times, one-dimensional.
            E : (n,) ArrayLike
                Vector of event types, entries evaluates as true for a recorded 
                event and false for censoring. 
            copy (bool):
                If true (default) parameters X, T, and E will be copied. Otherwise, 
                they will be copied only if they ar not tensors. 
        """
        def get_tensor(tensor):
            if copy:
                return tensor.clone()
            else:
                return tensor

        # Copy or convert X
        if torch.is_tensor(X):

            self.X = get_tensor(X)
                
        else:
            self.X = torch.tensor(X, dtype=torch.get_default_dtype())
        N = self.X.shape[0]

        # Copy or convert T to onedimensional tensor
        if torch.is_tensor(T):
            self.T = get_tensor(T)
        else:
            self.T = torch.tensor(T, dtype=torch.get_default_dtype())
        # Check dimensions
        if len(self.T.shape) > 1:
            warnings.warn("T not onedimensional, continuing by reshaping it.")
            self.T = self.T.reshape(-1,)
        # Check size
        if not self.T.shape[0] == N:
            raise ValueError(
                "Number of elements in T not equal to shape of first dimension of X")

         # Copy or convert E
        if E is None:
            self.E = torch.ones_like(self.T, dtype=torch.bool)
        else:
            if torch.is_tensor(E):
                self.E = get_tensor(E)
            else:
                self.E = torch.tensor(E)
            # Check type
            if not self.E.dtype == torch.bool:
                warnings.warn("E not bool, continuing by converting it.")
                self.E = self.E.type(torch.bool)
            # Check dimensions
            if len(self.E.shape) > 1:
                warnings.warn(
                    "E not onedimensional, continuing by reshaping it.")
                self.E = self.E.reshape(-1,)
            # Check size
            if not self.E.shape[0] == N:
                raise ValueError(
                    "Number of elements in E not equal to shape of first dimension of X")

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        return self.X[idx], self.T[idx], self.E[idx]

    def right_censor(self, censoring_time: float):
        r"""
        Right censors the dataset.

        Parameters
        ----------
            censoring_time (float):
                Censoring time

        Returns
        ----------
            None
        """
        self.E[self.T > censoring_time] = False
        self.T[self.T > censoring_time] = censoring_time

    def save_to_file(self, filename):
        r"""
        Saves dataset to a file.

        Parameters
        ----------
            filename (str):
                Name of file where dataset will be saved.

        Returns
        ----------
            None
        """
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load_from_file(cls, filename: str) -> Dataset:
        r"""
        Loads a dataset from a file.

        Parameters
        ----------
            filename (str): 
                Name of file from where to load the dataset.

        Returns
        ----------
            Dataset
        """
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def split(self, proportions: ArrayLike = None, sizes = None, generator=None):


        if proportions:
            proportions = np.array(proportions, ndmin=1).reshape(-1,)

            sizes = (proportions*len(self)).round().astype(int)
            if sizes.sum()<len(self):
                sizes = np.concatenate((sizes,[len(self)-sizes.sum()]))
        elif not sizes:
            raise ValueError("Either proportions or sizes must be provided")


        return torch.utils.data.random_split(self, sizes, generator=generator)



        # N = self.__len__()

        # I = np.array(range(N))
        # np.random.shuffle(I)

        # if not isinstance(x, list):
        #     x = [x]

        # s = (np.rint(np.array(x) * N).astype(np.int64))
        # r = N - s.sum()
        # s = s.cumsum()
        # s = list(s)

        # I_s = self._split_ind(I, s)

        # datasets = []
        # for n in range(len(I_s)):
        #     datasets.append(
        #         Dataset(self.X[I_s[n], :], self.T[I_s[n]], self.E[I_s[n]]))

        # return datasets

    @staticmethod
    def _split_ind(X, I):
        Xs = [X[:I[0]]]
        for n in range(1, len(I)):
            Xs.append(X[I[n - 1]:I[n]])
        Xs.append(X[I[-1]:])
        return Xs

    def __repr__(self):
        return (f"PySaRe Dataset with:\n" +
                f"  {self.T.shape[0]} individuals\n" +
                f"  {self.T.shape[0]-self.E.sum()} censored\n" +
                f"  {tuple(self.X[0].shape)} as shape of each individual's features")


    def plot(self):
        import lifelines

        km_event = lifelines.KaplanMeierFitter().fit(self.T.numpy(),self.E.numpy())
        ax = km_event.plot()
        km_cens = lifelines.KaplanMeierFitter().fit(self.T.numpy(),~self.E.numpy())
        km_cens.plot(ax=ax)

        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles=handles, labels=['Events', 'Censoring'])
        ax.set_xlabel('Time')
        ax.set_ylabel('Survival function')
        
        
    def to(self, device:Optional[DeviceLikeType] = None, dtype:Optional[_dtype]=None, data_transform:Optional[Callable]=None):
        r"""
        Performs Tensor dtype and/or device conversion. 

        Parameters
        ----------
            device (DeviceLikeType):
                Device to move dataset to. 
            dtype (_dtype):
                dtype to change the dataset to. 
            data_transform (Callable):
                Function that performs the descried conversion of X, T, and E like
                X_new, T_new, E_new = data_transform(X, T, E)
        Returns
        ----------
            Dataset:
                reference to self.
        """
        if data_transform is None:
            args = {arg: val for arg, val in (
                ("device", device), ("dtype", dtype)) if val is not None}

            self.X = self.X.to(**args)
            self.T = self.T.to(**args)

            # dtype of E should no be changed
            args.pop("dtype", None)
            self.E = self.E.to(**args)
            
        else:
            self.X, self.T, self.E = data_transform(self.X, self.T, self.E)

        return self


from typing import Optional, Union, Iterable, Callable, Literal, TypeVar, NewType, Any


Batch = NewType('Batch', Any)
AxisLike = NewType('AxisLike', Any)

X = NewType('X', Any)
T = NewType('T', Any)
E = NewType('E', Any)


import math
import torch
from torch import Tensor
from pysare.models.mixture._MixtureSurvivalModel import MixtureSurvivalModel
from pysare.models.utils.MLP import MLP
from ._mixers._Mixer import SoftmaxMixerWithReflection
from typing import Callable

sqrt2 = math.sqrt(2)
sqrt2pi = math.sqrt(2*math.pi)


class NormalForMixture(torch.nn.Module):

    num_parameters = 2

    def __init__(self, mu_activation=torch.exp, sigma_activation=torch.exp) -> None:
        super().__init__()
        self.mu_activation = mu_activation
        self.sigma_activation = sigma_activation

    def _get_parameters(self, parameters, T):
        Nd = parameters.shape[1]//2
        Nt = len(T)

        T = T.view(-1, 1).expand(Nt, Nd)
        mu = self.mu_activation(parameters[:, ::2])
        sigma = self.sigma_activation(parameters[:, 1::2])

        return mu, sigma, T

    def survival_probability(self, parameters, T):
        mu, sigma, T = self._get_parameters(parameters, T)
        return 0.5 * (1 - torch.erf((T - mu) / (sigma*sqrt2)))

    def lifetime_density(self, parameters, T):
        mu, sigma, T = self._get_parameters(parameters, T)
        return torch.exp(-0.5*((T-mu)/sigma)**2)/(sigma*sqrt2pi)
    
    def forward(self):
        return None


class GaussianMixture(MixtureSurvivalModel):
    r"""Gaussian mixture model with reflection.
    
    Implements a mixture of reflected Gaussian components. Each component is 
    parameterized by a location and scale parameter, corresponding to the mean 
    and standard deviation of the underling gaussian distribution. 

    .. figure:: ../figures/reflected_kernel.pdf
        :class: with-border
        :align: center
        :width: 500px

        Illustration of how a component is reflected around time zero.

    Parameters
    ----------
    location_activation : callable, default=torch.exp
        Activation function for the location parameter. Takes a tensor and
        returns a tensor of the same shape and with non-negative elements.
    scale_activation : callable, default=torch.exp
        Activation function for the scale parameter. Takes a tensor and
        returns a tensor of the same shape and with non-negative elements.
    """
    
    def __init__(self, 
                 location_activation: Callable[[Tensor], Tensor] =torch.exp, 
                 scale_activation: Callable[[Tensor], Tensor] =torch.exp) -> None:
        
        super().__init__(distributions=NormalForMixture(mu_activation=location_activation, 
                                                        sigma_activation=scale_activation),
                         mixer=SoftmaxMixerWithReflection())
    
    def forward_output_size(self, num_components: int):
        return num_components*3

    @classmethod
    def MLP_implementation(cls, num_components, input_size, hidden_sizes,
                           activation=torch.nn.ReLU, dropout=0.0, batch_norm=False,
                           shape_activation=torch.exp, scale_activation=torch.exp):
        return MLPNormalMixture(num_components, input_size, hidden_sizes,
                                activation, dropout, batch_norm,
                                shape_activation, scale_activation)


class MLPNormalMixture(GaussianMixture):
    def __init__(self, num_components, input_size, hidden_sizes,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False,
                 location_activation=torch.exp, scale_activation=torch.exp) -> None:
        super().__init__(location_activation=location_activation,
                         scale_activation=scale_activation)

        self.layers = MLP(input_size, num_components*3,
                          hidden_sizes, activation, dropout, batch_norm)
        self.num_components = num_components

    def forward(self, X):
        return self.layers(X)



import torch
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models.piecewise.grid import DiscretizationGrid
from abc import abstractmethod
from pysare.models.utils import MLP

class LinearHazard(TorchSurvivalModel):
    r"""Implements an piecewise linear hazard model
    
    The forward output size should be 
        number of intervals + 1 + potential extra inputs required by the grid

    See :mod:`pysare.models.piecewise` for information on how to use 
    piecewise models.
    """
    def __init__(self, grid: DiscretizationGrid):
        super(LinearHazard, self).__init__()
        self.grid = grid

    def _log_likelihood(self, z, T, E):
        N = T.shape[0]
        disc, z = self.grid.bin(T, z)

        e = torch.exp(z)
        dH = (e[:, :-1] + e[:, 1:])*disc.dt/2

        H_n = dH.cumsum(dim=1)

        h = e[range(N), disc.ind] + (T-disc.tau_ind) * \
            (e[range(N), disc.ind+1] - e[range(N), disc.ind])/disc.dt_ind

        H = H_n[range(N), disc.ind] - (disc.tau_next_ind-T) * \
            e[range(N), disc.ind+1] - (disc.tau_next_ind-T)**2 / \
            (2*disc.dt_ind)*(e[range(N), disc.ind] - e[range(N), disc.ind+1])

        # S = torch.exp(-H)

        # f = h*S

        ind_c = torch.where(~E)[0]
        ln_f = torch.log(h) - H
        ln_S = -H
        log_likelihood = ln_f
        log_likelihood[ind_c] = ln_S[ind_c]

        return log_likelihood

    def max_time(self):
        return self.grid.max_time()
    
    def forward_output_size(self):
        return self.grid.num_intervals()+self.grid.extra_forward_output_size()+1
    
    def _density_function(self, z, T):
        N = T.shape[0]
        disc,z = self.grid.bin(T,z)

        e = torch.exp(z)
        dH = (e[:, :-1] + e[:, 1:])*disc.dt/2

        H_n = dH.cumsum(dim=1)

        h = e[range(N), disc.ind] + (T-disc.tau_ind) * \
            (e[range(N), disc.ind+1] - e[range(N), disc.ind])/disc.dt_ind

        H = H_n[range(N), disc.ind] - (disc.tau_next_ind-T) * \
            e[range(N), disc.ind+1] - (disc.tau_next_ind-T)**2 / \
            (2*disc.dt_ind)*(e[range(N), disc.ind] - e[range(N), disc.ind+1])

        S = torch.exp(-H)
        # f = h*S
        return h*S

    def _survival_function(self, z, T):
        N = T.shape[0]
        disc, z = self.grid.bin(T, z)

        e = torch.exp(z)
        dH = (e[:, :-1] + e[:, 1:])*disc.dt/2

        H_n = dH.cumsum(dim=1)

        h = e[range(N), disc.ind] + (T-disc.tau_ind) * \
            (e[range(N), disc.ind+1] - e[range(N), disc.ind])/disc.dt_ind

        H = H_n[range(N), disc.ind] - (disc.tau_next_ind-T) * \
            e[range(N), disc.ind+1] - (disc.tau_next_ind-T)**2 / \
            (2*disc.dt_ind)*(e[range(N), disc.ind] - e[range(N), disc.ind+1])

        S = torch.exp(-H)
        return S

    def forward(self, X):
        return X

    @classmethod
    def MLP_implementation(cls, grid, input_size, hidden_sizes,
                           activation=torch.nn.ReLU, dropout=False,
                           batch_norm=False):
        return MLPLinearHazard(grid, input_size, hidden_sizes,
                               activation, dropout,
                               batch_norm)


class MLPLinearHazard(LinearHazard):
    def __init__(self, grid, input_size, hidden_sizes,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False):
        super(MLPLinearHazard, self).__init__(grid=grid)

        self.input_size = input_size

        self.layers = MLP(input_size, grid.num_intervals() + grid.extra_forward_output_size() + 1,
                          hidden_sizes, activation, dropout, batch_norm)

    def forward(self, X):
        return self.layers(X)



import time
import torch
import pysare
import numpy as np
import matplotlib.pyplot as plt
import torch.utils.data
import pysare
import numpy as np
import matplotlib.pyplot as plt
import lifelines as lf
import pandas as pd
from collections.abc import Iterable
from matplotlib.axes._axes import Axes
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models._numpy.NumpySurvivalModel import NumpySurvivalModel
from pysare.data import Dataset
from torch.utils.data.dataloader import DataLoader
from typing import Union, Optional, Tuple, Literal, NewType
from numpy.typing import ArrayLike
from pandas import DataFrame

from pysare._typing import X, T, E


def negative_log_likelihood(
    model: Union[TorchSurvivalModel, NumpySurvivalModel],
        data: Union[Tuple[X, T, E], Dataset, DataLoader],
        plot: Union[bool, Axes, Iterable[Axes]] = False
) -> DataFrame:
    r"""Negative log-likelihood of a model on a dataset

    Calculates the truncated negative log-likelihood for a model on a dataset. 
    Truncated refers to that the mean log-likelihood of all event times up to 
    a specific time is calculated, the last value thus corresponds to the 
    non-truncated value.

    Parameters
    ----------
    
        model : SurvivalModel or NumpySurvivalModel
            Model to be evaluated. If data is a tuple (X, T, E) of non-tensors, 
            the model should be a NumpySurvivalModel, and otherwise it should 
            be a SurvivalModel.
        data : tuple(X,T,E) or Dataset or Dataloader
            The data to be used for evaluation. If data is a tuple (X, T, E) of 
            non-tensors, the model should be a NumpySurvivalModel, and otherwise 
            it should be a SurvivalModel.

        plot : bool or Axis, default=False
            If True, a plot of the result will be created in a new figure; if 
            an Axes is passed, the plots will be made in this.

    Returns
    ----------

        result : DataFrame
            A pandas DataFrame with the result.
    """


    if hasattr(model, 'eval'):
        model.eval()

    # Interpret inputs
    # ==========================================================================
    if isinstance(data, tuple):  # data is a tuple (X, T, E)
        dataloader = (data,)
    elif isinstance(data, Iterable):  # data is a DataLoader
        dataloader = data
    else:  # Data is a Dataset
        dataloader = (data[:],)

    dataset = data

    # Full vector of times and log-likelihoods
    # ===========================================================================
    T = []
    E = []
    l = []
    with torch.no_grad():
        if isinstance(model, pysare.models.TorchSurvivalModel):
            for X_batch, T_batch, E_batch in dataloader:
                T.append(np.array(T_batch.reshape(-1,), copy=True))
                l.append(np.array(model.log_likelihood(
                    X_batch, T_batch, E_batch), copy=True))
        else:
            for X_batch, T_batch, E_batch in dataloader:
                    T.append(np.array(T_batch.reshape(-1,), copy=True))
                    l.append(np.array(model.log_likelihood(
                        X_batch, T_batch.reshape(-1,1), E_batch.reshape(-1,1)), 
                        copy=True))

    T = np.concatenate(T)
    l = np.concatenate(l)

    order = T.argsort()
    T = T[order]
    l = l[order]

    nll = -l.cumsum()/np.arange(1, len(l)+1)


    # Plot if ax is not False
    if plot:

        if plot is True:
            fig, ax = plt.subplots(nrows=1)
        else:
            ax = plot

        # ax.annotate(f'End: {nll[-1]:6.6g}', (T[-1], nll[-1]),
        #      textcoords="offset points", xytext=(-100,0),
        #      ha='center',va='center', arrowprops=dict(arrowstyle="->"))

        ax.plot(T, nll)
        ax.set_ylabel('Truncated negative log-likelihood')
        ax.set_xlabel('Truncation Time')
        
        if plot is True:
            plt.tight_layout()

    return pd.DataFrame(data={'negative_log_lik': nll, 'time': T}).set_index('time')



import torch
import matplotlib.pyplot as plt
import torch_lr_finder


def negative_log_likelihood(model, batch):
    return -model.log_likelihood(*batch)


class WrapLoss(torch.nn.Module):
    r"""A Wrapper used to wrap the loss for lr_finder
    """

    def __init__(self, model, loss_function=negative_log_likelihood, loss_aggregation='mean') -> None:
        super().__init__()

        self.model = model
        self.loss_function = loss_function

        if loss_aggregation == 'mean':
            self.loss_aggregation = torch.mean
        elif loss_aggregation == 'sum':
            self.loss_aggregation = torch.sum
        else:
            raise ValueError("loss_aggregation must be 'mean' or 'sum'")

    def forward(self, batch: torch.Tensor, _: torch.Tensor) -> torch.Tensor:

        model = self.model
        loss = self.loss_function(self.model, batch)

        return self.loss_aggregation(loss)


class WrapModel(torch.nn.Module):
    r"""A Wrapper used to wrap a model for lr_finder
    """

    def __init__(self, model) -> None:
        super().__init__()
        self.model = model

    def forward(self, batch):
        return batch


class WrapDataLoaders(object):
    r"""A Wrapper used to change the collate function of a datalaoder
    in lr_finder"""

    def __init__(self, train_loader, val_loader=None):

        self.original_train_collate_fn = train_loader.collate_fn
        self.train_loader = train_loader

        self.val_loader = val_loader
        if val_loader:
            self.original_val_collate_fn = val_loader.collate_fn

    def __enter__(self):

        self.original_train_collate_fn = self.train_loader.collate_fn

        def train_collate_fn(batch):
            return self.original_train_collate_fn(batch), None

        self.train_loader.collate_fn = train_collate_fn

        if self.val_loader:
            def val_collate_fn(batch):
                return self.original_train_collate_fn(batch), None

            self.original_val_collate_fn = self.val_loader.collate_fn
            self.val_loader.collate_fn = val_collate_fn

        return self.train_loader, self.val_loader

    def __exit__(self, *args):
        self.train_loader.collate_fn = self.original_train_collate_fn
        if self.val_loader:
            self.val_loader.collate_fn = self.original_val_collate_fn


class LRFinder:
    r"""Learning rate range test based on the package torch_lr_finder. 

    The learning rate range test increases the learning rate in a pre-training 
    run between two boundaries in a linear or exponential manner. The result 
    can be used to determine an appropriate learning rate, see references below.

    Arguments:
    ----------

        model (Module): 
            Model.
        optimizer (torch.optim.Optimizer): 
            Optimizer where the defined learning is assumed to be the lower 
            boundary of the range test.
        loss_function (torch.nn.Module): 
            wrapped loss function.
        memory_cache (bool): 
            If this flag is set to True, `state_dict` of model and optimizer 
            will be cached in memory. Otherwise, they will be saved to files 
            under the `cache_dir`.
        cache_dir (string): 
            path for storing temporary files. If no path is specified, 
            system-wide temporary directory is used (default). Notice that this
            parameter will be ignored if `memory_cache` is True.

    Examples:
    ---------

        >>> lr_finder = LRFinder(model, optimizer)
        >>> lr_finder.range_test(dataloader, end_lr=100, num_iter=100, reset=False)
        >>> lr_finder.plot() # to inspect the loss-learning rate graph
        >>> lr_finder.reset() # to reset the model and optimizer to their initial 
                              # state, done automatically if reset=True above

    
    References:
    -----------

        Cyclical Learning Rates for Training Neural Networks: 
            https://arxiv.org/abs/1506.01186
        fastai/lr_find: 
            https://github.com/fastai/fastai
    """

    def __init__(self, model, optimizer, loss_function=negative_log_likelihood,
                 loss_aggregation='mean', memory_cache=True, cache_dir=None) -> None:

        self.backend_lr_finder = torch_lr_finder.LRFinder(model=WrapModel(model),
                                                          optimizer=optimizer,
                                                          criterion=WrapLoss(
                                                              model, loss_function, loss_aggregation),
                                                          memory_cache=memory_cache,
                                                          cache_dir=cache_dir)
        self.suggested_lr = None

    def range_test(
        self,
        train_loader,
        val_loader=None,
        start_lr=None,
        end_lr=10,
        num_iter=100,
        step_mode="exp",
        smooth_f=0.05,
        diverge_th=5,
        accumulation_steps=1,
        non_blocking_transfer=True,
        reset=True
    ):
        r"""Performs the learning rate range test. 

        Arguments:
        -------
            train_loader (Iterable[Batch]):
                The training set data loader. The dataloader should iterate 
                over batches that are used as inputs to loss_function.
            val_loader (Iterable[Batch]): 
                When given a data loader, the model is evaluated after each 
                iteration on that dataset and the evaluation loss is used. Note 
                that in this mode the test takes significantly longer, but 
                generally produces more precise results. Default: None.
            start_lr (float): 
                The starting learning rate for the range test.
                Default: None (uses the learning rate from the optimizer).
            end_lr (float): 
                The maximum learning rate to test. Default: 10.
            num_iter (int): 
                The number of iterations over which the test occurs. 
                Default: 100.
            step_mode (str): 
                One of the available learning rate policies, linear or 
                exponential ("linear", "exp"). Default: "exp".
            smooth_f (float): 
                The loss smoothing factor within the [0, 1[ interval. 
                Disabled if set to 0, otherwise the loss is smoothed using
                exponential smoothing. Default: 0.05.
            diverge_th (int): 
                The test is stopped when the loss surpasses the threshold:  
                diverge_th * best_loss. Default: 5.
            accumulation_steps (int): 
                steps for gradient accumulation. If it is 1, gradients are not 
                accumulated. Default: 1.
            non_blocking_transfer (bool): 
                when non_blocking_transfer is set, tries to convert/move data 
                to the device asynchronously if possible, e.g., moving CPU 
                Tensors with pinned memory to CUDA devices. Default: True.
            reset (bool): 
                If True, the the model and optimizer is restored to  their 
                initial states. Default: True.

        Examples:
        -------
        Example (fastai approach):
            >>> lr_finder = LRFinder(net, optimizer, criterion, device="cuda")
            >>> lr_finder.range_test(dataloader, end_lr=100, num_iter=100)

        Example (Leslie Smith's approach):
            >>> lr_finder = LRFinder(net, optimizer, criterion, device="cuda")
            >>> lr_finder.range_test(trainloader, val_loader=val_loader, end_lr=1, num_iter=100, step_mode="linear")

        Gradient accumulation is supported; example:
            >>> train_data = ...    # prepared dataset
            >>> desired_bs, real_bs = 32, 4         # batch size
            >>> accumulation_steps = desired_bs // real_bs     # required steps for accumulation
            >>> dataloader = torch.utils.data.DataLoader(train_data, batch_size=real_bs, shuffle=True)
            >>> acc_lr_finder = LRFinder(net, optimizer, criterion, device="cuda")
            >>> acc_lr_finder.range_test(dataloader, end_lr=10, num_iter=100, accumulation_steps=accumulation_steps)

        Reference:
        [Training Neural Nets on Larger Batches: Practical Tips for 1-GPU, Multi-GPU & Distributed setups](
        https://medium.com/huggingface/ec88c3e51255)
        [thomwolf/gradient_accumulation](https://gist.github.com/thomwolf/ac7a7da6b1888c2eeac8ac8b9b05d3d3)
        """
        with WrapDataLoaders(train_loader, val_loader) as (wraped_train_loader, wraped_val_loader):
            # range_test currently has no return, but for possible future
            # compatibility it is assumed it might have
            result = self.backend_lr_finder.range_test(
                train_loader,
                val_loader,
                start_lr,
                end_lr,
                num_iter,
                step_mode,
                smooth_f,
                diverge_th,
                accumulation_steps,
                non_blocking_transfer)
        if reset:
            self.reset()
        return result

    def plot(self,
             skip_start=10,
             skip_end=5,
             log_lr=True,
             show_lr=None,
             ax=None,
             suggest_lr=True):
        r"""Plots the learning rate range test.

        Arguments:
        ----------
            skip_start (int): number of batches to trim from the start.
                Default: 10.
            skip_end (int): number of batches to trim from the start.
                Default: 5.
            log_lr (bool): True to plot the learning rate in a logarithmic
                scale; otherwise, plotted in a linear scale. Default: True.
            show_lr (float): if set, adds a vertical line to visualize the
                specified learning rate. Default: None.
            ax (matplotlib.axes.Axes): the plot is created in the specified
                matplotlib axes object and the figure is not be shown. If `None`, 
                then the figure and axes object are created in this method and 
                the figure is shown . Default: None.
            suggest_lr (bool): 
                If True, suggest a learning rate by 'steepest': the point with 
                steepest gradient (minimal gradient). You can use that point as 
                a first guess for an LR. Default: True.

        Returns:
        --------
            The matplotlib.axes.Axes object that contains the plot,
            and the suggested learning rate (if set suggest_lr=True).
        """
        result = self.backend_lr_finder.plot(skip_start,
                                             skip_end,
                                             log_lr,
                                             show_lr,
                                             ax,
                                             suggest_lr)
        if suggest_lr:
            self.suggested_lr = result[1]

        return result

    def set_lr(self, lr='suggested'):
        r"""Sets the learning rate of the optimizer

        Arguments:
        ----------
            lr ('suggested', float): 
            learning rate to be set. If 'suggested', the suggested learning 
            rate is used, the method plot must have been run with argument 
            suggest_lr True before this. If float, this value is set.
        """

        if lr == 'suggested':
            if self.suggested_lr:
                lr = self.suggested_lr
            else:
                raise ValueError(('Suggested learning rate not calculated.',
                                  'run method plot  with argument suggest_lr True.'))

        for group in self.backend_lr_finder.optimizer.param_groups:
            group['lr'] = lr

    def reset(self):
        r"""Restores the model and optimizer to their initial states. Is run 
        by default in range_test if parameter reset i not set to false."""
        return self.backend_lr_finder.reset()



import numpy as np
import torch
from torch import nn
import torch
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models.utils import MLP
from pysare.models.energy_based.integrators._Integrator import EnergyIntegrator, EquidistantTrapezoidal, RandomMonteCarlo
from torch.nn import Module
from typing import Optional, Union

class EnergyBasedDensity(TorchSurvivalModel):
    r"""Implements an energy based survival model.

    In this model, the density is directly specified using a normalized
    neural network. The unnormalized output from the network is called the 
    energy, hence the name. See [1] for more information.

    The model uses two networks, one in the forward function and one to
    calculate the energy:

    - Forward network
        This network is implemented in the forward function 
        and takes the covariates/features X an should return a tensor of shape 
        `(batch_size, energy_input_size-1)` where batch_size is number of 
        samples in the batch (size of first dimension of X) and
        `energy_input_size` is the input size of the energy network.
        By default the forward network implements the identity function (that is, 
        `forward(X)=X`).
    - Energy network
        The energy network is used to calculate the energy based
        on the output from the  forward network. The input to the energy network
        is the output from the forward network concatenated with a tensor of the 
        times of interest. The output from the energy network should be a tensor of 
        shape (batch_size,1) with numbers in (-inf, inf). 

    This model can either be subclassed or instanced. If instanced, a network 
    must be passed as the parameter energy_net and the forward function will
    be the identity function (that is, `forward(X)=X`).

    The normalization of the energy network is done using an integrator 
    from :mod:`pysare.models.energy_based.integrators`. The model only 
    explicitly model the density up to a specified maximal time, specified
    by the parameter `max_time`. The remaining tail is modeled using a single 
    point to after `max_time`, whose location is determined by the parameter
    `tail_ratio` as `tail_ratio*max_time`. See the figure below for an 
    illustration.  

    .. figure:: ../figures/tail_illustration.pdf
        :class: with-border
        :align: center
        :width: 500px

        Illustration of how the tail is treated in the integration. The blue 
        area is approximated using numerical integration and the tail is 
        modeled as the red area, calculated using a single point. Here :math:`t_m` 
        corresponds to `max_time` and :math:`\gamma` to 
        `tail_ratio`. See [1] for details.

    Parameters:
    -----------

    max_time : float
        Maximal prediction time. 
    tail_ratio : float, default=1.2
        Float that determines with point is used to calculate the integral of 
        the tail of the energy network. The tail is approximated as 
        `tail_energy * max_time * (1- tail_ratio)`
        where `tail_energy` is the energy at time `max_time*tail_ratio`. 
    train_integrator : EnergyIntegrator or int
        Integrator used during training. Set as active integrator after 
        construction and automatically set when the model is set to training mode 
        using the `train` method. Can be changed using the `set_integrators` 
        method. If and integer is provided, `RandomMonteCarlo` with the 
        specified number of samples is used.
    eval_integrator : EnergyIntegrator or int
        Integrator used during evaluation. Automatically set when the model is 
        set to evaluation mode using the `eval` method. Can be changed using 
        the `set_integrators` method. If and integer is provided, 
        `EquidistantTrapezoidal` with the specified number of samples 
        is used.
    energy_net : Module, optional
        If supplied this net will be used as energy. The `energy_net.forward` 
        method should take a tensor of shape `(batch_size, forward_out+1)` and 
        return a tensor of shape `(batch_size, 1)`, where 
        `(batch_size, forward_out)` is the shape of the output from the 
        `self.forward` method when the batch size is `batch_size`; the "+1"
        in the input shape comes from that the energy also has the time as an
        input. If the `self.forward` method is not specified `forward_out` 
        is the size of the covariate/feature vector.

    References
    ----------

    [1] Holmer, Olov, Erik Frisk, and Mattias Krysander. "Energy-Based Survival
    Models for Predictive Maintenance." IFAC-PapersOnLine 56.2 (2023): 10862-10867.
    https://doi.org/10.48550/arXiv.2302.00629

    """

    def __init__(self, max_time:float,  
                 train_integrator: Union[EnergyIntegrator, int], 
                 eval_integrator: Union[EnergyIntegrator, int], 
                 tail_ratio:float=1.2, 
                 energy_net: Optional[Module]=None):
        
        super().__init__()

        self._max_time = max_time
        self.tail_ratio = tail_ratio

        if isinstance(train_integrator, EnergyIntegrator):
            self._train_integrator = train_integrator
        else:
            self._train_integrator = RandomMonteCarlo(int(train_integrator))

        if isinstance(eval_integrator, EnergyIntegrator):
            self._eval_integrator = eval_integrator
        else:
            self._eval_integrator = EquidistantTrapezoidal(int(eval_integrator))
            
        self._active_integrator = train_integrator

        if energy_net is not None:
            self._energy_net = energy_net
        else:
            self._energy_net = None

    def train(self, mode: bool = True):
        super().train(mode)
        if mode:
            self._active_integrator = self._train_integrator
        else:
            self._active_integrator = self._eval_integrator
        return self

    def eval(self):
        return self.train(False)
    
    def max_time(self):
        return self._max_time

    def forward_output_size(self):
        return 1

    def _log_likelihood(self, forward_out, T, E):

        log_Z0 = self._active_integrator.log_integrate(
            self, forward_out, torch.zeros((T.shape[0], 1)), self._max_time, self.tail_ratio)

        log_Z = self._active_integrator.log_integrate(
            self, forward_out, T, self._max_time, self.tail_ratio)

        z = self.energy(torch.cat((forward_out, T.reshape(-1, 1)), dim=1))

        l = - log_Z0
        l[E] += z[E]
        l[~E] += log_Z[~E]
        return l

    def _density_function(self, forward_out, T):
        Z = self._active_integrator.integrate(
            self, forward_out, torch.zeros((T.shape[0], 1)), self._max_time, self.tail_ratio)

        f = torch.exp(self.energy(torch.cat((forward_out, T.reshape(-1, 1)), dim=1))) \
            / Z.reshape(-1, 1)
        return f

    def _survival_function(self, forward_out, T):
        Z0 = self._active_integrator.integrate(
            self, forward_out, torch.zeros((T.shape[0], 1)), self._max_time, self.tail_ratio)
        Z = self._active_integrator.integrate(
            self, forward_out, T, self._max_time, self.tail_ratio)
        return Z/Z0

    def energy(self, X_T):
        return self._energy_net(X_T)
    
    def forward(self, X):
        return X

    @classmethod
    def MLP_implementation(cls, max_time, train_integrator, eval_integrator,
                           input_size, hidden_sizes,
                           tail_ratio=1.2,
                           activation=torch.nn.ReLU, dropout=False,
                           batch_norm=False):

        return MLP_EBD(max_time,  train_integrator, eval_integrator,
                       input_size, hidden_sizes,tail_ratio,
                       activation, dropout,
                       batch_norm)


class MLP_EBD(EnergyBasedDensity):
    def __init__(self, max_time, train_integrator, eval_integrator,
                 input_size, hiden_sizes, tail_ratio=1.2,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False):
        

        super(MLP_EBD, self).__init__(max_time, train_integrator, eval_integrator,
                                      tail_ratio)
        
        self.layers = MLP(input_size=input_size+1,
                          output_size=1,
                          hidden_sizes=hiden_sizes,
                          activation=activation,
                          dropout=dropout,
                          batch_norm=batch_norm)
        self.input_size = input_size

    def energy(self, X_T):
        return self.layers(X_T)




import torch
import os
import tempfile
import copy

class StateCacher:
    def __init__(self, in_memory:bool =False, path:str=None, overwrite:bool = False):

        self.in_memory = in_memory

        if in_memory:
            self.memmory_cache = None
            self.delete=False
        else:
            if path is None:
                self.path = tempfile.gettempdir()+'temp.ckpt'
                self.delete = True
            else:
                self.path = path
                if (not overwrite) and (not os.path.isfile(path)):
                    raise ValueError(f"File {path} already exists.")
                self.delete = False
        
        self.empty = True

    def save(self, module):
        if self.in_memory:
            self.memmory_cache = copy.deepcopy(module.state_dict())
        else:
            torch.save(module.state_dict(), self.path)
        self.empty = False

    def load(self, module):
        if self.in_memory:
            module.load_state_dict(self.memmory_cache)
        else:
            module.load_state_dict(torch.load(self.path))

    def delete_temp(self):
        r"""Deletes temporary files"""
        if self.delete and os.path.exists(self.path):
            os.remove(self.path)
        self.empty = True

    def __del__(self):
        r"""Checks if there is a file to delete"""
        self.delete_temp()

        


r"""Integrators to be used in energy based models.

The following integrators are available:

.. autosummary:: 
    :toctree: generated
    :recursive:

    RandomMonteCarlo
    EquidistantTrapezoidal
    AdaptiveTrapezoidalSimpsons

"""

from ._Integrator import RandomMonteCarlo
from ._Integrator import EquidistantTrapezoidal
from ._Integrator import AdaptiveTrapezoidalSimpsons


import torch
from pysare.models.SurvivalModel import TorchSurvivalModel


class Uniform(TorchSurvivalModel):
    r"""Uniform distribution, parameters=[low, high]

    Implements a uniform distribution with parameters [low, high] (output from 
    forward function) where low and high are the lower respective upper bound 
    of the support of the distribution.

    """

    def __init__(self) -> None:
        super().__init__()

    def max_time(self):
        return torch.inf

    def _inv_survival_function(self, forward_out, S):
        return forward_out[:, 0]+(1.0-S)*(forward_out[:, 1]-forward_out[:, 0])

    def inv_survival_function(self, X, S):
        T = torch.full_like(S, torch.nan)
        inds = (S >= 0.0) & (S <= 1.0)
        if inds.any():
            T[inds] = self._inv_survival_function(
                self.forward(X[inds]), S[inds])
        return T

    def _survival_function(self, forward_out, t):
        return (1.0-(t-forward_out[:, 0])/(forward_out[:, 1] -
                                           forward_out[:, 0])).clip(0.0, 1.0)

    def _density_function(self, forward_out, t):
        return ((t >= forward_out[:, 0]) & (t <= forward_out[:, 1]))/(forward_out[:, 1] - forward_out[:, 0])

    def _sample(self, forward_out, num_samples=1, generator=None):
        u = torch.rand((forward_out.shape[0], num_samples),
                       device=forward_out.device,
                       dtype=forward_out.dtype,
                       generator=generator)
        return forward_out[:, [0]] + u*(forward_out[:, [1]]-forward_out[:, [0]])

    def forward(self, X):
        return X



import setuptools


description = "PySaRe -- Neural Network-Based Survival Analysis and Reliability Engineering"

with open("README.md") as f:
    long_description, long_description_content_type = f.read(), "text/markdown"

setuptools.setup(
    name="pysare",
    version="0.1.2",
    author="Olov Holmer",
    author_email="olov.holmer@liu.se",
    description=description,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    url="https://github.com/oholmer/PySaRe",
    project_urls={
        "Bug Tracker": "https://github.com/oholmer/PySaRe/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.10",
    install_requires=["numpy>=1.14.0", "torch>=0.4.1","pandas>=2.1","matplotlib>=3.0","torch_lr_finder >=0.2.1","scipy>=1.7.0","lifelines>=0.29.0"],
)


from torch.nn import Module
from torch import Tensor
import torch
from ._numpy.NumpySurvivalModel import NumpySurvivalModel

from typing import Optional, Union, Iterable, Callable, Literal, TypeVar, NewType, Any, Tuple
from pysare._typing import X, T, E
from pysare.data import Dataset
from torch.utils.data import DataLoader


class TorchSurvivalModel(Module):
    r"""Base class for survival models.

    This class is subclassed in all torch-based survival models.
    """

    def __init__(self):
        super(TorchSurvivalModel, self).__init__()

    def max_time(self) -> float:
        r"""Returns the maximal time for which the model is defined"""
        raise NotImplementedError()

    def forward_output_size(self, *args, **kwargs):
        r"""Returns the required size of the forward output."""
        raise NotImplementedError()

    def to_numpy(self,
                 data: Optional[Union[Tuple[X, T, E],
                                      Dataset,
                                      DataLoader]] = None,
                 torch_max_time: Optional[float] = None,
                 numpy_support: Optional[Tuple[float, float]] = None,
                 batch_size: Optional[int] = None,
                 extrapolation: Literal['set_zero', 'model'] = 'set_zero') -> NumpySurvivalModel:
        r"""Converts the model to a NumpySurvivalModel that accepts numpy inputs.

        Implements an interface to a PyTorch model creating a survival model that
        takes NumPy arrays (or any object that can be converted to such) as input.


        Parameters
        ----------

        data : tuple(X, T, E) or Dataset or DataLoader, optional
            Data used to initialize the interface. If data should represent the
            tensors X, T, and E, and the device and dtype of these tensors
            are used to determine the correct tensor types for the torch_model.
            If data is a DataLoader its batch size will also be used in the
            interface, if not batch_size is set.
        torch_max_time : float, optional
            Defines the maximal time of the original model before extrapolation 
            is to be used. By default, self.max_time() is used.
        numpy_support : tuple(float, float), optional
            Support of the constructed NumPy model. The numpy model will map
            [0.0, torch_max_time] onto numpy_support using an affine
            transformation. By default the same support as the original model 
            is used.
        batch_size : int, optional
            Maximal batch size to be used when calling the underlying PyTorch
            Model. By default an infinite batch size, or the size of the
            DataLoader optionally specified as the data parameter.
        extrapolation : {'set_zero', 'model'}, default='set_zero'
            Determines how the model is extrapolated beyond torch_max_time.
            - 'set_zero' -- Survival function is set to zero
            - 'model' -- The value given by the torch_model is used even beyond
                its maximal time. Can result in unexpected behavior and should be
                used with caution.
        """
        return NumpySurvivalModel(self,
                                  data=data,
                                  torch_max_time=torch_max_time,
                                  numpy_support=numpy_support,
                                  batch_size=batch_size,
                                  extrapolation=extrapolation)

    def sample(self, X: Tensor, num_samples: int = 1, generator: Optional[torch.Generator] = None, **kwargs):
        with torch.no_grad():
            return self._sample(self.forward(X), num_samples, generator, **kwargs)

    def _sample(self, forward_out, num_samples=1, generator=None, burn_in=20, shuffle=True):
        r"""Metropolis-Hastings sampling

        Default sampling method for models that does not have a specific 
        implementation.

        The Metropolis-Hastings algorithm is used to sample from the
        distribution of (T | T<max_time). After the Metropolis-Hastings 
        algorithm, each sample is set to max_time with probability
        1-S(max_time).

        """

        N = forward_out.shape[0]

        if not self.max_time() < torch.inf:
            raise NotImplementedError("This model has nonfinite max time" +
                                      "and can therefore not be sampled " +
                                      "using Metropolis-Hastings ")
        else:
            max_time = self.max_time()
            S_max_time = self._survival_function(
                forward_out,
                torch.full((N,),
                           max_time,
                           device=forward_out.device,
                           dtype=forward_out.dtype))

        T = torch.empty((N, num_samples),
                        device=forward_out.device,
                        dtype=forward_out.dtype)

        T_last = torch.rand((N,),
                            device=forward_out.device,
                            dtype=forward_out.dtype,
                            generator=generator)*max_time

        if burn_in < 1:
            T[:, 0] = T_last
            # Tail of the distribution
            T[torch.rand(T_last.shape, dtype=T_last.dtype,
                         device=T_last.device, generator=generator)
              < S_max_time, 0] = max_time

        f_last = self._density_function(forward_out, T_last)

        for n in range(1, num_samples+burn_in):
            proposal = torch.rand((N,),
                                  device=forward_out.device,
                                  dtype=forward_out.dtype,
                                  generator=generator)*max_time

            f_new = self._density_function(forward_out, proposal)

            accept = torch.rand((N,),
                                device=forward_out.device,
                                dtype=forward_out.dtype,
                                generator=generator) <= f_new/f_last

            if accept.any():
                T_last[accept] = proposal[accept]
                f_last[accept] = f_new[accept]

            if n >= burn_in:
                T[:, n-burn_in] = T_last
                # Tail of the distribution
                T[torch.rand(T_last.shape, dtype=T_last.dtype,
                             device=T_last.device, generator=generator)
                  < S_max_time, n-burn_in] = max_time

        if shuffle:
            shuffling_inds = torch.argsort(torch.rand(T.shape,
                                                      dtype=T.dtype,
                                                      device=T.device,
                                                      generator=generator),
                                           dim=1)
            return torch.gather(T, dim=1, index=shuffling_inds)
        else:
            return T

    def log_likelihood(self, X: Tensor, T: Tensor, E: Tensor) -> Tensor:
        r"""Returns the likelihood of the observation (X,T,E)

        Parameters
        ----------

            X : (n, ...) tensor
                Tensor with features/covariate with first dimension as batch 
                dimension 
            T : (n, ) tensor
                Tensor with times
            E : (n,) tensor
                Tensor with event types
        """
        Z = self.forward(X)
        return self._log_likelihood(Z, T, E)

    def _log_likelihood(self, Z: Tensor, T: Tensor, E: Tensor) -> Tensor:

        log_likelihood = torch.empty_like(T)

        if torch.any(E):
            log_likelihood[E] = self._log_density_function(Z[E], T[E])

        if torch.any(~E):
            log_likelihood[~E] = self._log_survival_function(Z[~E], T[~E])

        return log_likelihood

    def density_function(self, X: Tensor, T: Tensor) -> Tensor:
        r"""Lifetime density at T given X

        Parameters
        ----------

            X : (n, ...) tensor
                Tensor with features/covariate with first dimension as batch 
                dimension 
            T : (n, ) tensor
                Tensor with times

        """
        return self._density_function(self.forward(X), T)

    def _density_function(self, Z: Tensor, T: Tensor) -> Tensor:
        raise NotImplementedError()

    def hazard_function(self, X: Tensor, T: Tensor) -> Tensor:
        r"""Hazard function at at T given X

           Parameters
        ----------

            X : (n, ...) tensor
                Tensor with features/covariate with first dimension as batch 
                dimension 
            T : (n, ) tensor
                Tensor with times

        """
        return self._hazard_function(self.forward(X), T)

    def _hazard_function(self, Z: Tensor, T: Tensor) -> Tensor:
        return self._density_function(Z, T)/self._survival_function(Z, T)

    def log_density_function(self, X: Tensor, T: Tensor) -> Tensor:
        r"""Logarithm of the density function at T given X

         Parameters
        ----------

            X : (n, ...) tensor
                Tensor with features/covariate with first dimension as batch 
                dimension 
            T : (n, ) tensor
                Tensor with times

        """
        return self._log_density_function(self.forward(X), T)

    def _log_density_function(self, Z: Tensor, T: Tensor) -> Tensor:
        return torch.log(self._density_function(Z, T))

    def survival_function(self, X: Tensor, T: Tensor) -> Tensor:
        r"""Survival function at T given X

         Parameters
        ----------

            X : (n, ...) tensor
                Tensor with features/covariate with first dimension as batch 
                dimension 
            T : (n, ) tensor
                Tensor with times

        """
        return self._survival_function(self.forward(X), T)

    def _survival_function(self, Z: Tensor, T: Tensor) -> Tensor:
        raise NotImplementedError()

    def log_survival_function(self, X: Tensor, T: Tensor) -> Tensor:
        r"""Logarithm of the survival function at T given X

         Parameters
        ----------

            X : (n, ...) tensor
                Tensor with features/covariate with first dimension as batch 
                dimension 
            T : (n, ) tensor
                Tensor with times
        """
        return self._log_survival_function(self.forward(X), T)

    def _log_survival_function(self, Z: Tensor, T: Tensor) -> Tensor:
        return torch.log(self._survival_function(Z, T))


class SplineSurvivalModel(TorchSurvivalModel):
    def __init__(self, times, probabilities):
        super(SplineSurvivalModel, self).__init__()

        self.times = torch.nn.Parameter(torch.cat((times, torch.tensor(
            [2*times[-1]-times[-2]], dtype=times.dtype, device=times.device)), dim=-1), requires_grad=False)
        self.probabilities = torch.nn.Parameter(torch.cat((probabilities, torch.tensor(
            [0.0], dtype=probabilities.dtype, device=probabilities.device)), dim=-1), requires_grad=False)
        self.dt = torch.nn.Parameter(self.times.diff(), requires_grad=False)
        self.density = torch.nn.Parameter(-self.probabilities.diff() /
                                          self.dt, requires_grad=False)

    def sample(self, X, generator=None):

        u = torch.rand(X.shape[0], generator=generator)

        # Search sorted requires an increasing sequence, so CDF probabilities
        # are used
        P = 1-self.probabilities
        ind = torch.searchsorted(P, u, right=True)-1
        ind[u == 1-self.probabilities[-1]] -= 1

        time = self.times[ind]
        density = self.density[ind]
        probability = P[ind]

        return (time + (u-probability)/density).clip(0, self.max_time())

    def max_time(self):
        return self.times[-2]

    def _find_ind(self, T):
        T = T.clip(0, self.times[-1])
        ind = torch.searchsorted(self.times, T, right=True)-1
        ind[T == self.times[-1]] = len(self.times)-2
        return ind

    def _density_function(self, X, T, batch_size=None) -> Tensor:
        return self.density[self._find_ind(T)]

    def _survival_function(self, X, T, batch_size=None) -> Tensor:
        ind = self._find_ind(T)

        density = self.density[ind]
        probability = self.probabilities[ind]

        dt = self.dt[ind]
        time = self.times[ind]

        return probability-density*(T-time)

    def to_numpy(self, dtype=None, device=None):
        return NumpySurvivalModel(self, dtype, device, )


class FrozenSurvivalModel(Module):
    def __init__(self, base_model, X):
        super(FrozenSurvivalModel, self).__init__()

        self.base_model = base_model
        self.X = X

    def _feature_vector(self, N):
        return self.X.expand(N, *self.X.shape)

    def max_time(self):
        return self.base_model.max_time()

    def sample(self, size=1, generator=None):
        return self.base_model.sample(self._feature_vector(size), generator)

    def log_likelihood(self, T, E, batch_size=None) -> Tensor:
        return self.base_model.log_likelihood(self._feature_vector(len(T)), T.reshape(-1,), E.reshape(-1,))

    def lifetime_density(self, T, batch_size=None) -> Tensor:
        return self.base_model.density_function(self._feature_vector(len(T)), T.reshape(-1,))

    def survival_function(self, T, batch_size=None) -> Tensor:

        return self.base_model.survival_function(self._feature_vector(len(T)), T.reshape(-1,))

    def to_numpy(self, dtype=None, device=None):
        return NumpyFrozenSurvivalModel(self, dtype, device)

    def to_spline_approximation(self, discretization):
        if torch.is_tensor(discretization) and (discretization.numel() > 1):
            times = discretization
        else:
            times = torch.linspace(
                0, self.max_time(), int(discretization), dtype=self.X.dtype, device=self.X.device)

        probabilities = self.survival_function(times)

        return FrozenSurvivalModel(SplineSurvivalModel(times, probabilities), self.X)



r"""Piecewise defined models.

The piecewise models use a discretization grid from 
:mod:`pysare.models.piecewise.grid` and and a neural network 
to specify a piecewise model.  

The following models are available:

- :class:`ConstantDensity`: 
    Piecewise constant density (num_intervals+1)
- :class:`LinearDensity`: 
    Piecewise linear density (num_intervals+2)
- :class:`ConstantHazard`: 
    Piecewise constant hazard (num_intervals)
- :class:`LinearHazard`: 
    Piecewise linear hazard (num_intervals+1)

where the number in parenthesis is the required size of the output
from the forward method, if the chosen grid does not require extra
outputs. The correct output size can also be determined using
the method forward_output_size().

Below is an example of a :class:`LinearDensity` model with a 
:class:`~grid.EquidistantGrid`:

    >>> from pysare.models.piecewise import LinearDensity
    >>> from pysare.models.piecewise.grid import UniformGrid
    ...
    >>> class Model(LinearDensity):
    ...     def __init__(self):
    ...         super().__init__(grid=UniformGrid(max_time, num_intervals))
    ...         self.net = ...
    ...         self.final_layer = torch.nn.Linear(...,
    ...                                       self.forward_output_size())
    ...
    ...     def forward(self, X):
    ...         return self.final_layer(self.net(X))

"""

from ._ConstantDensity import ConstantDensity
from ._LinearDensity import LinearDensity
from ._ConstantHazard import ConstantHazard
from ._LinearHazard import LinearHazard
from . import grid




import time

class Printer():
    def __init__(self, print_time) -> None:

        self.print_time = print_time

        self.last_print = time.time()-self.print_time

        self.last_train_loss = f"{'-':>12}"
        self.last_validation_loss = f"{'-':>12}"

    def finalize(self):
        print("")
        print("-"*57)

    def loss_str(self, loss):
        if loss:
            return f"{loss:>12.6g}"
        else:
            return f"{'-':>12}"

    def int_str(self, num):
        if num < 1000000:
            return f"{num:6.0f}"
        else:
            return f"{num:>2.2g}"

    def train_update(self, epoch, batch, batch_loss):
        self.batch_loss = self.loss_str(batch_loss)
        self.epoch = self.int_str(epoch)
        self.batch = self.int_str(batch)

        self.print()
    
    def print(self):
        pass

class SingleLinePrinter(Printer):
    def __init__(self, print_time) -> None:
        super().__init__(print_time)

        print(f"{'Current batch:':<26} | {'Previous epoch:':<27}")
        print(f"{'Epoch':>6} {'Batch':>6} {'Train. Loss':>12} | {'Train. loss':>12}{'':>3}{'Valid. loss':>12}")
        print("-"*57)

    def finalize_update(self, epoch, batch, train_loss, validation_loss, batch_loss):
        self.last_train_loss = self.loss_str(train_loss)
        self.last_validation_loss = self.loss_str(validation_loss)
        self.batch_loss = self.loss_str(batch_loss)
        self.epoch = self.int_str(epoch)
        self.batch = self.int_str(batch)

        self.print()

    def print(self):
        if time.time()-self.last_print > self.print_time:
            print(self.epoch + " " + self.batch + " " + self.batch_loss + " | " +
                  self.last_train_loss + "   "+self.last_validation_loss, end='\r')
            self.last_print = time.time()


class MultipleLinePrinter(Printer):

    def __init__(self, print_time, print_new_line_count=1) -> None:
        super().__init__(print_time)

        self.print_new_line_count = print_new_line_count
        self.skip_count = 0

        print(f"{'Current batch:':<26} | {'Full epoch:':<27}")
        print(f"{'Epoch':>6} {'Batch':>6} {'Train. Loss':>12} | {'Train. loss':>12}{'':>3}{'Valid. loss':>12}")
        print("-"*57)

        self.last_train_loss = f"{'-':>12}"
        self.last_validation_loss = f"{'-':>12}"
        self.print_finalize = False

    def finalize_update(self, epoch, batch, train_loss, validation_loss, batch_loss):
        self.last_train_loss = self.loss_str(train_loss)
        self.last_validation_loss = self.loss_str(validation_loss)
        self.batch_loss = self.loss_str(batch_loss)
        self.epoch = self.int_str(epoch)
        self.batch = self.int_str(batch)
        self.print(finalize=True)
        self.last_train_loss = f"{'-':>12}"
        self.last_validation_loss = f"{'-':>12}"

    def _print(self, end='\r'):
        print(self.epoch + " " + self.batch + " " + self.batch_loss + " | " +
              self.last_train_loss + "   "+self.last_validation_loss, end=end)

    def print(self, finalize=False):
        if finalize:

            self.skip_count += 1

            if self.skip_count >= self.print_new_line_count:
                if self.print_finalize or (time.time()-self.last_print > self.print_time):
                    self._print("\n")
                    self.skip_count = 0
                    self.last_print = time.time()
                    self.print_finalize = False
                    return
            self.print_finalize = False
        if time.time()-self.last_print > self.print_time:
            self._print()
            self.last_print = time.time()
            self.print_finalize = True




import torch
from pysare.models.mixture._MixtureSurvivalModel import MixtureSurvivalModel
from pysare.models.utils.MLP import MLP
from ._mixers._Mixer import SoftmaxMixer
from typing import Callable
from torch import Tensor



class WeibullForMixture(torch.nn.Module):

    num_parameters = 2

    def __init__(self, shape_activation=torch.exp, scale_activation=torch.exp) -> None:
        super().__init__()
        self.shape_activation = shape_activation
        self.scale_activation = scale_activation

    def _get_parameters(self, parameters, T):
        Np = 2
        Nd = parameters.shape[1]//2
        Nt = len(T)

        T = T.view(-1, 1).expand(Nt, Nd)
        shape = self.shape_activation(parameters[:, ::2])
        scale = self.scale_activation(parameters[:, 1::2])

        return shape, scale, T

    def survival_probability(self, parameters, T):
        shape, scale, T = self._get_parameters(parameters, T)
        return torch.exp(- (T/scale)**shape)

    def lifetime_density(self, parameters, T):
        shape, scale, T = self._get_parameters(parameters, T)
        return (shape/scale)*(T/scale)**(shape-1)*torch.exp(- (T/scale)**shape)


class WeibullMixture(MixtureSurvivalModel):
    r"""Weibull mixture model.
    
    Implements a mixture of Weibull components. Each component is parameterized 
    by a shape and scale parameter, and have the survival function 
    `S(t) = exp((t/scale)**shape)`
     

    For a model with M components, each row in the forward function should be 
    on the form 
    `[shape_1, scale_1, ..., shape_M, scale_M, weight_1, ... weight_M]`
    where shape_i, scale_i are the shape and scale parameters for component i, 
    before activation; and weight_1,... weight_M are the weights before 
    activation.

    Parameters
    ----------

    shape_activation : callable, default=torch.exp
        Activation function for the shape parameter. Takes a tensor and
        returns a tensor of the same shape and with non-negative elements.
    scale_activation : callable, default=torch.exp
        Activation function for the scale parameter. Takes a tensor and
        returns a tensor of the same shape and with non-negative elements.
    """
    def __init__(self, 
                 shape_activation:Callable[[Tensor], Tensor]=torch.exp, 
                 scale_activation:Callable[[Tensor], Tensor]=torch.exp) -> None:

        super().__init__(distributions=WeibullForMixture(
            shape_activation=torch.exp, scale_activation=torch.exp), mixer=SoftmaxMixer())


    @classmethod
    def MLP_implementation(cls, num_components, input_size, hidden_sizes,
                           activation=torch.nn.ReLU, dropout=0.0, batch_norm=False,
                           shape_activation=torch.exp, scale_activation=torch.exp):
        return MLPWeibullMixture(num_components, input_size, hidden_sizes,
                                 activation, dropout, batch_norm,
                                 shape_activation, scale_activation)

    def forward_output_size(self, num_components: int):
        return num_components*3

class MLPWeibullMixture(WeibullMixture):
    def __init__(self, num_components, input_size, hidden_sizes,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False,
                 shape_activation=torch.exp, scale_activation=torch.exp) -> None:
        super().__init__(shape_activation=shape_activation,
                         scale_activation=scale_activation)

        self.layers = MLP(input_size, num_components*3,
                          hidden_sizes, activation, dropout, batch_norm)
        self.num_components = num_components

    def forward(self, X):
        return self.layers(X)

# %%


r"""Implementations and tools for defining survival models. 

For an introduction on how to define and use models, see 
https://github.com/oholmer/PySaRe/blob/main/Introduction.ipynb.

Neural Network-Based Models
---------------------------
The following neural network-based models are available as submodules:

-  :mod:`pysare.models.piecewise` -- Piecewise defined models
-  :mod:`pysare.models.mixture` -- Mixture models
-  :mod:`pysare.models.energy_based` -- Energy based models
-  :mod:`pysare.models.distributions` -- Standard distributions


Model Functionality
-------------------     
PySaRe models extend PyTorch modules to provide the following methods:

- :py:meth:`SurvivalModel.survival_probability`
- :py:meth:`SurvivalModel.lifetime_density`
- :py:meth:`SurvivalModel.hazard_function`
- :py:meth:`SurvivalModel.log_likelihood`


NumPy Interface
---------------
PyTorch module requires the inputs to be a tensor of correct dtype and device, 
and to simply the usage of models a model can be converted to handle any inputs
that can be converted to numpy Arrays. This is done using the :func:`to_numpy`
method:

    >>> np_model = model.to_numpy()

Frozen Models
-------------
Predictions are often done multiple times for the same covariates/features and 
a model can therefore be frozen at a specific covariate creating a frozen model 
that does not require the covariate as input. This is done using the 
:meth:`feeze` method:      

    >>> frozen_model = model.freeze(X[0])


Importing Models
----------------

- :func:`pysare.model.from_survival_probabilities`
    Creates a model from a list of times and corresponding survival probabilities.
- :func:`pysare.model.from_survival_function`
    Creates a spline approximation of a given survival function.

The following example creates an approximation of a Weibull distribution:
    
    >>> model = pysare.models.from_survival_function(lambda t: math.exp(-t**2))  

Frozen models can also be created directly from a list of survival probabilities
and can thus be used as a way to use models from other sources in PySaRe.

    >>> frozen_model = pysare.models.from_survival_probabilities(times, probs)

"""

from . import _numpy
from . import piecewise
from . import distributions
from . import energy_based
from . import mixture
from . import utils

from .ensemble import ensemble

from ._KaplanMeier import KaplanMeierEstimator

from .SurvivalModel import TorchSurvivalModel
# from . import energy_based





import lifelines
import torch
import numpy as np

def QQ_plot(model, dataloader, num_times=np.inf, model_opts='--+'):

    # Full vector of times  and events
    # ===========================================================================
    T = []
    E = []

    for X_batch, T_batch, E_batch in dataloader:
        T.append(T_batch.reshape(-1,))
        E.append(E_batch.reshape(-1,))

    if torch.is_tensor(T[0]):
        T = torch.concat(T)
        E = torch.concat(E)
        T_np = np.array(T.cpu())
        E_np = np.array(E.cpu())
    else:
        T = np.concatenate(T)
        E = np.concatenate(E)
        T_np = T
        E_np = E

    def find_nearest(array, values):
        inds = np.searchsorted(array, values).clip(1)
        inds[ values - array[inds-1] < array[inds] - values] -= 1
        return inds

    order = np.argsort(T_np)
    T, E, T_np, E_np = T[order], E[order], T_np[order], E_np[order]



    T_event, T_event_np = T[E], T_np[E_np]

    if num_times < len(T_event):
        t = np.linspace(T_event_np[0], T_event_np[-1], num_times)

        inds = np.unique(find_nearest(T_event_np, t))
        T_event, T_event_np = T_event[inds], T_event_np[inds]


    KM = lifelines.KaplanMeierFitter().fit(T_np, E_np)
    ax = KM.plot()


    S_np = np.empty_like(T_event_np)

    with torch.no_grad():
        for n in range(T_event_np.shape[0]):

            R = []
            for X_batch, _, _ in dataloader:
                R.append(np.array(model.survival_function(
                    X_batch, T_event[n:n+1].repeat(X_batch.shape[0])).reshape(-1,).cpu(), copy=False))
            R = np.concatenate(R)

            S_np[n] = R.mean()

    ax.plot(T_event_np, S_np,model_opts, label='Model mean over covariates')
    ax.legend()

    handles, _ = ax.get_legend_handles_labels()
    ax.legend(handles=handles, labels=['Kaplan-Meier estimate', 'Model mean over covariates'])
    ax.set_xlabel('Time')
    ax.set_ylabel('Survival function')
